pub struct AtemPacket {
    length: u16,
    flags: u8,
    session_id: u16,
    remote_packet_id: u16,
    body: Vec<u8>,
}

pub enum AtemPacketErr {
    TooShort(String),
    LengthDiffers(String),
}

#[derive(PartialEq)]
pub enum PacketFlag {
    AckRequest,
    NewSessionId,
    IsRetransmit,
    RetransmitRequest,
    AckReply,
}

impl From<PacketFlag> for u8 {
    fn from(flag: PacketFlag) -> Self {
        match flag {
            PacketFlag::AckRequest => 0x01,
            PacketFlag::NewSessionId => 0x02,
            PacketFlag::IsRetransmit => 0x04,
            PacketFlag::RetransmitRequest => 0x08,
            PacketFlag::AckReply => 0x10,
        }
    }
}

impl AtemPacket {
    pub fn length(&self) -> u16 {
        self.length
    }

    pub fn flags(&self) -> u8 {
        self.flags
    }

    pub fn session_id(&self) -> u16 {
        self.session_id
    }

    pub fn remote_packet_id(&self) -> u16 {
        self.remote_packet_id
    }

    pub fn body(&self) -> Vec<u8> {
        self.body.clone()
    }

    pub fn has_flag(&self, flag: PacketFlag) -> bool {
        self.flags & u8::from(flag) > 0
    }
}

impl TryFrom<&[u8]> for AtemPacket {
    type Error = AtemPacketErr;

    fn try_from(buffer: &[u8]) -> Result<Self, Self::Error> {
        if buffer.len() < 12 {
            return Err(AtemPacketErr::TooShort(format!(
                "Invalid packet from ATEM {:x?}",
                buffer
            )));
        }

        let length = u16::from_be_bytes(buffer[0..2].try_into().unwrap()) & 0x07ff;
        if length as usize != buffer.len() {
            return Err(AtemPacketErr::LengthDiffers(format!(
                "Length of message differs, expected {} got {}",
                length,
                buffer.len()
            )));
        }

        let flags = buffer[0] >> 3;
        let session_id = u16::from_be_bytes(buffer[2..4].try_into().unwrap());
        let remote_packet_id = u16::from_be_bytes(buffer[10..12].try_into().unwrap());

        let body = buffer[12..].to_vec();

        Ok(AtemPacket {
            length,
            flags,
            session_id,
            remote_packet_id,
            body,
        })
    }
}
