use std::{
    io,
    net::SocketAddr,
    time::{Duration, SystemTime},
};

use log::debug;
use tokio::net::UdpSocket;

use crate::atem_lib::atem_util;

const IN_FLIGHT_TIMEOUT: u64 = 60;
const CONNECTION_TIMEOUT: u64 = 5000;
const CONNECTION_RETRY_INTERVAL: u64 = 1000;
const RETRANSMIT_CHECK_INTERVAL: u64 = 1000;
const MAX_PACKET_RETRIES: u16 = 10;
const MAX_PACKET_ID: u16 = 1 << 15;
const MAX_PACKET_PER_ACK: u16 = 16;

// Set to max UDP packet size, for now
const MAX_PACKET_RECEIVE_SIZE: usize = 65535;
const ACK_PACKET_LENGTH: u16 = 12;

#[derive(PartialEq, Clone)]
enum ConnectionState {
    Closed,
    SynSent,
    Established,
}

#[allow(clippy::from_over_into)]
impl Into<u8> for ConnectionState {
    fn into(self) -> u8 {
        match self {
            ConnectionState::Closed => 0x00,
            ConnectionState::SynSent => 0x01,
            ConnectionState::Established => 0x02,
        }
    }
}

#[derive(PartialEq)]
enum PacketFlag {
    AckRequest,
    NewSessionId,
    IsRetransmit,
    RetransmitRequest,
    AckReply,
}

impl From<PacketFlag> for u8 {
    fn from(flag: PacketFlag) -> Self {
        match flag {
            PacketFlag::AckRequest => 0x01,
            PacketFlag::NewSessionId => 0x02,
            PacketFlag::IsRetransmit => 0x04,
            PacketFlag::RetransmitRequest => 0x08,
            PacketFlag::AckReply => 0x10,
        }
    }
}

#[derive(Clone)]
struct InFlightPacket {
    packet_id: u16,
    tracking_id: u64,
    payload: Vec<u8>,
    pub last_sent: SystemTime,
    pub resent: u16,
}

struct AckedPacket {
    packet_id: u16,
    tracking_id: u64,
}

pub struct AtemSocketCommand {
    payload: Vec<u8>,
    raw_name: String,
    tracking_id: u64,
}

pub struct AtemSocketInner {
    connection_state: ConnectionState,
    reconnect_timer: Option<SystemTime>,
    retransmit_timer: Option<SystemTime>,

    next_send_packet_id: u16,
    session_id: u16,

    socket: Option<UdpSocket>,
    address: String,
    port: u16,

    last_received_at: SystemTime,
    last_received_packed_id: u16,
    in_flight: Vec<InFlightPacket>,
    ack_timer: Option<SystemTime>,
    received_without_ack: u16,
}

enum AtemSocketReceiveError {
    Closed,
}

#[derive(Debug, Error)]
enum AtemSocketWriteError {
    #[error("Socket closed")]
    Closed,

    #[error("Socket disconnected")]
    Disconnected(#[from] io::Error),
}

impl AtemSocketInner {
    pub fn new() -> Self {
        AtemSocketInner {
            connection_state: ConnectionState::Closed,
            reconnect_timer: None,
            retransmit_timer: None,

            next_send_packet_id: 1,
            session_id: 0,

            socket: None,
            address: "0.0.0.0".to_string(),
            port: 0,

            last_received_at: SystemTime::now(),
            last_received_packed_id: 0,
            in_flight: vec![],
            ack_timer: None,
            received_without_ack: 0,
        }
    }

    pub async fn connect(&mut self, address: String, port: u16) -> Result<(), io::Error> {
        self.address = address.clone();
        self.port = port;

        let socket = UdpSocket::bind("0.0.0.0:0").await?;
        let remote_addr = format!("{}:{}", address, port)
            .parse::<SocketAddr>()
            .unwrap();
        socket.connect(remote_addr).await?;
        self.socket = Some(socket);

        self.start_timers();

        self.next_send_packet_id = 1;
        self.session_id = 0;
        self.in_flight = vec![];
        debug!("Reconnect");

        self.send_packet(&atem_util::COMMAND_CONNECT_HELLO).await;
        self.connection_state = ConnectionState::SynSent;

        Ok(())
    }

    pub fn disconnect(&mut self) {
        self.stop_timers();

        self.retransmit_timer = None;
        self.reconnect_timer = None;
        self.ack_timer = None;
        self.socket = None;

        let prev_connection_state = self.connection_state.clone();
        self.connection_state = ConnectionState::Closed;

        if prev_connection_state == ConnectionState::Established {
            self.on_disconnect();
        }
    }

    pub async fn send_commands(&mut self, commands: Vec<AtemSocketCommand>) {
        for command in commands.into_iter() {
            self.send_command(&command.payload, &command.raw_name, command.tracking_id)
                .await;
        }
    }

    pub async fn send_command(&mut self, payload: &[u8], raw_name: &str, tracking_id: u64) {
        let packet_id = self.next_send_packet_id;
        self.next_send_packet_id += 1;
        if self.next_send_packet_id >= MAX_PACKET_ID {
            self.next_send_packet_id = 0;
        }

        let opcode = u16::from(u8::from(PacketFlag::AckRequest)) << 11;

        let mut buffer = vec![0; 20 + payload.len()];

        // Headers
        buffer[0..2].copy_from_slice(&u16::to_be_bytes(opcode | (payload.len() as u16 + 20)));
        buffer[2..4].copy_from_slice(&u16::to_be_bytes(self.session_id));
        buffer[10..12].copy_from_slice(&u16::to_be_bytes(packet_id));

        // Command
        buffer[12..14].copy_from_slice(&u16::to_be_bytes(payload.len() as u16 + 8));
        buffer[16..20].copy_from_slice(raw_name.as_bytes());

        // Body
        buffer[20..20 + payload.len()].copy_from_slice(payload);
        self.send_packet(&buffer).await;

        debug!("{:x?}", buffer);

        self.in_flight.push(InFlightPacket {
            packet_id,
            tracking_id,
            payload: buffer,
            last_sent: SystemTime::now(),
            resent: 0,
        })
    }

    async fn restart_connection(&mut self) {
        self.disconnect();
        self.connect(self.address.clone(), self.port).await.ok();
    }

    pub async fn tick(&mut self) {
        let messages = self.receive().await.ok();
        if let Some(messages) = messages {
            for message in messages.iter() {
                self.recieved_packet(message).await;
            }
        }
        if let Some(ack_time) = self.ack_timer {
            if ack_time <= SystemTime::now() {
                self.ack_timer = None;
                self.received_without_ack = 0;
                self.send_ack(self.last_received_packed_id).await;
            }
        }
        if let Some(reconnect_time) = self.reconnect_timer {
            if reconnect_time <= SystemTime::now() {
                if self.last_received_at + Duration::from_millis(CONNECTION_TIMEOUT)
                    <= SystemTime::now()
                {
                    debug!("{:?}", self.last_received_at);
                    debug!("Connection timed out, restarting");
                    self.restart_connection().await;
                }
                self.start_reconnect_timer();
            }
        }
        if let Some(retransmit_time) = self.retransmit_timer {
            if retransmit_time <= SystemTime::now() {
                self.check_for_retransmit().await;
                self.start_retransmit_timer();
            }
        }
    }

    async fn receive(&mut self) -> Result<Vec<Vec<u8>>, AtemSocketReceiveError> {
        let mut messages: Vec<Vec<u8>> = vec![];
        let socket = self.socket.as_mut().ok_or(AtemSocketReceiveError::Closed)?;

        let mut buf = [0; MAX_PACKET_RECEIVE_SIZE];
        if let Ok((message_size, _)) = socket.try_recv_from(&mut buf) {
            messages.push(buf[0..message_size].to_owned());
        }

        Ok(messages)
    }

    fn is_packet_covered_by_ack(&self, ack_id: u16, packet_id: u16) -> bool {
        let tolerance: u16 = MAX_PACKET_ID / 2;
        let pkt_is_shortly_before = packet_id < ack_id && packet_id + tolerance > ack_id;
        let pkt_is_shortly_after = packet_id > ack_id && packet_id < ack_id + tolerance;
        let pkt_is_before_wrap = packet_id > ack_id + tolerance;
        packet_id == ack_id
            || ((pkt_is_shortly_before || pkt_is_before_wrap) && !pkt_is_shortly_after)
    }

    async fn recieved_packet(&mut self, packet: &[u8]) {
        debug!("RECV {:x?}", packet);

        if packet.len() < 12 {
            debug!("Invalid packet from ATEM {:x?}", packet);
            return;
        }

        self.last_received_at = SystemTime::now();
        let length = u16::from_be_bytes(packet[0..2].try_into().unwrap()) & 0x07ff;

        if length as usize != packet.len() {
            debug!(
                "Length of message differs, expected {} got {}",
                length,
                packet.len()
            );
            return;
        }

        let flags = packet[0] >> 3;
        self.session_id = u16::from_be_bytes(packet[2..4].try_into().unwrap());
        let remote_packet_id = u16::from_be_bytes(packet[10..12].try_into().unwrap());

        if flags & u8::from(PacketFlag::NewSessionId) > 0 {
            self.connection_state = ConnectionState::Established;
            self.last_received_packed_id = remote_packet_id;
            self.send_ack(remote_packet_id).await;
            return;
        }

        if self.connection_state == ConnectionState::Established {
            if flags & u8::from(PacketFlag::RetransmitRequest) > 0 {
                let from_packet_id = u16::from_be_bytes(packet[6..8].try_into().unwrap());
                debug!("Retransmit request: {:x?}", from_packet_id);

                self.retransmit_from(from_packet_id).await;
            }

            if flags & u8::from(PacketFlag::AckRequest) > 0 {
                if remote_packet_id == (self.last_received_packed_id + 1) % MAX_PACKET_ID {
                    self.last_received_packed_id = remote_packet_id;
                    self.send_or_queue_ack().await;

                    if length > 12 {
                        self.on_command_received(&packet[12..], remote_packet_id);
                    }
                } else if self
                    .is_packet_covered_by_ack(self.last_received_packed_id, remote_packet_id)
                {
                    self.send_or_queue_ack().await;
                }
            }

            if flags & u8::from(PacketFlag::IsRetransmit) > 0 {
                debug!("ATEM retransmitted packet {:x?}", remote_packet_id);
            }

            if flags & u8::from(PacketFlag::AckReply) > 0 {
                let ack_packet_id = u16::from_be_bytes(packet[4..6].try_into().unwrap());
                let mut acked_commands: Vec<AckedPacket> = vec![];

                self.in_flight = self
                    .in_flight
                    .clone()
                    .into_iter()
                    .filter(|pkt| {
                        if self.is_packet_covered_by_ack(ack_packet_id, pkt.packet_id) {
                            acked_commands.push(AckedPacket {
                                packet_id: pkt.packet_id,
                                tracking_id: pkt.tracking_id,
                            });
                            false
                        } else {
                            true
                        }
                    })
                    .collect();
                self.on_command_acknowledged(acked_commands);
            }
        }
    }

    async fn send_packet(&self, packet: &[u8]) {
        debug!("Send {:x?}", packet);
        if let Some(socket) = &self.socket {
            socket.send(packet).await.ok();
        } else {
            debug!("Socket is not open")
        }
    }

    async fn send_or_queue_ack(&mut self) {
        self.received_without_ack += 1;
        if self.received_without_ack >= MAX_PACKET_PER_ACK {
            self.received_without_ack = 0;
            self.ack_timer = None;
            self.send_ack(self.last_received_packed_id).await;
        } else if self.ack_timer.is_none() {
            self.ack_timer = Some(SystemTime::now() + Duration::from_millis(5));
        }
    }

    async fn send_ack(&mut self, packet_id: u16) {
        debug!("Sending ack for packet {:x?}", packet_id);
        let flag: u8 = PacketFlag::AckReply.into();
        let opcode = u16::from(flag) << 11;
        let mut buffer: [u8; ACK_PACKET_LENGTH as _] = [0; 12];
        buffer[0..2].copy_from_slice(&u16::to_be_bytes(opcode as u16 | ACK_PACKET_LENGTH));
        buffer[2..4].copy_from_slice(&u16::to_be_bytes(self.session_id));
        buffer[4..6].copy_from_slice(&u16::to_be_bytes(packet_id));
        self.send_packet(&buffer).await;
    }

    async fn retransmit_from(&mut self, from_id: u16) {
        let from_id = from_id % MAX_PACKET_ID;

        if let Some(index) = self
            .in_flight
            .iter()
            .position(|pkt| pkt.packet_id == from_id)
        {
            debug!(
                "Resending from {} to {}",
                from_id,
                self.in_flight[self.in_flight.len() - 1].packet_id
            );
            for i in index..self.in_flight.len() {
                let mut sent_packet = self.in_flight[i].clone();
                if sent_packet.packet_id == from_id
                    || !self.is_packet_covered_by_ack(from_id, sent_packet.packet_id)
                {
                    sent_packet.last_sent = SystemTime::now();
                    sent_packet.resent += 1;

                    self.send_packet(&sent_packet.payload).await;
                }
            }
        } else {
            debug!("Unable to resend: {}", from_id);
            self.restart_connection().await;
        }
    }

    async fn check_for_retransmit(&mut self) {
        for sent_packet in self.in_flight.clone() {
            if sent_packet.last_sent + Duration::from_millis(IN_FLIGHT_TIMEOUT) < SystemTime::now()
            {
                if sent_packet.resent <= MAX_PACKET_RETRIES
                    && self
                        .is_packet_covered_by_ack(self.next_send_packet_id, sent_packet.packet_id)
                {
                    debug!("Retransmit from timeout: {}", sent_packet.packet_id);

                    self.retransmit_from(sent_packet.packet_id).await;
                } else {
                    debug!("Packet timed out: {}", sent_packet.packet_id);
                    self.restart_connection().await;
                }
            }
        }
    }

    fn on_command_received(&mut self, payload: &[u8], packet_id: u16) {
        // TODO: Emit some event
    }

    fn on_command_acknowledged(&mut self, ids: Vec<AckedPacket>) {
        // TODO: Emit some event
    }

    fn on_disconnect(&mut self) {
        // TODO: Emit some event
    }

    fn start_timers(&mut self) {
        self.start_reconnect_timer();
        self.start_retransmit_timer();
    }

    fn stop_timers(&mut self) {
        self.reconnect_timer = None;
        self.retransmit_timer = None;
    }

    fn start_reconnect_timer(&mut self) {
        self.reconnect_timer =
            Some(SystemTime::now() + Duration::from_millis(CONNECTION_RETRY_INTERVAL));
    }

    fn start_retransmit_timer(&mut self) {
        self.retransmit_timer =
            Some(SystemTime::now() + Duration::from_millis(RETRANSMIT_CHECK_INTERVAL));
    }
}
